use crate::config::Config;
use crate::indexes::{SerializableMessage, SerializableThread};
use serde::{Deserialize, Serialize};
use std::fs::File;
use std::io::{BufReader};
use warp::Filter;
use warp::http::StatusCode;
use crate::{create_folder_lar, delete_folder_lar, rename_folder_lar};
use crate::sonic::SearchSonic;
use crate::util::read_and_decompress_file;

pub async fn run_api() {
    let cors = warp::cors()
        .allow_any_origin()
        .allow_methods(vec!["GET", "POST"])
        .allow_headers(vec![warp::http::header::CONTENT_TYPE]);

    let routes = warp::path("folders")
        .and(warp::get())
        .and_then(get_folders_handle)
        .or(warp::path("sorted_threads_by_date")
            .and(warp::get())
            .and(warp::query::<GetSortedThreadsByDateQuery>())
            .and_then(sorted_threads_by_date_handle))
        .or(warp::path("get_thread_messages")
            .and(warp::get())
            .and(warp::query::<GetThreadMessagesQuery>())
            .and_then(get_thread_messages_handle))
        .or(warp::path("email")
            .and(warp::get())
            .and(warp::query::<GetEmailQuery>())
            .and_then(get_email_handle))
        .or(warp::path("search")
            .and(warp::get())
            .and(warp::query::<SearchQuery>())
            .and_then(search_handle))
        .or(warp::path("create_folder")
            .and(warp::post())
            .and(warp::body::json())
            .and_then(create_folder_handle))
        .or(warp::path("rename_folder")
            .and(warp::post())
            .and(warp::body::json())
            .and_then(rename_folder_handle))
        .or(warp::path("delete_folder")
            .and(warp::post())
            .and(warp::body::json())
            .and_then(delete_folder_handle))
        .with(cors);

    warp::serve(routes).run(([127, 0, 0, 1], 3030)).await; // TODO from config
}

pub async fn get_folders() -> String {
    let mut folders: Vec<String> = Vec::new();
    for entry in std::fs::read_dir(Config::global().out_dir.clone())
        .expect("could not read maildir")
        .filter_map(|m| m.ok())
    {
        let metadata = match entry.metadata() {
            Ok(metadata) => metadata,
            Err(_) => continue,
        };
        if !metadata.is_dir() {
            continue;
        }
        folders.push(entry.file_name().into_string().unwrap());
    }

    serde_json::to_string(&folders).unwrap()
}

pub async fn get_email(id: String) -> String {
    let messages_path = Config::global()
        .out_dir
        .clone()
        .join("messages.json");
    let messages_file = match File::open(messages_path) {
        Ok(file) => file,
        Err(_) => return String::new(),
    };
    let messages_reader = BufReader::new(messages_file);
    let messages: Vec<SerializableMessage> = match serde_json::from_reader(messages_reader)
    {
        Ok(messages) => messages,
        Err(_) => return String::new(),
    };
    let message = match messages.iter().find(|message| message.id == id){
        None => {return String::new()}
        Some(message) => {message}
    };
    
    let abs_path = Config::global().out_dir.clone()
        .join(message.list.clone())
        .join("messages")
        .join(message.hash.clone() + ".tar.gz");
    let mut content = match read_and_decompress_file(abs_path, message.hash.clone()) {
        Ok(content) => {
            String::from_utf8_lossy(&content).to_string()
        }
        Err(_) => String::new()
    };
    
    content
}

pub async fn get_thread_messages(id: u32) -> String {
    let threads_path = Config::global()
        .out_dir
        .clone()
        .join("threads.json");
    let messages_path = Config::global()
        .out_dir
        .clone()
        .join("messages.json");
    
    let threads_file = match File::open(threads_path) {
        Ok(file) => file,
        Err(_) => return serde_json::to_string(&Vec::<String>::new()).unwrap(),
    };
    let messages_file = match File::open(messages_path) {
        Ok(file) => file,
        Err(_) => return serde_json::to_string(&Vec::<SerializableMessage>::new()).unwrap(),
    };
    
    let threads_reader = BufReader::new(threads_file);
    let messages_reader = BufReader::new(messages_file);
    
    let threads: Vec<SerializableThread> = match serde_json::from_reader(threads_reader)
    {
        Ok(threads) => threads,
        Err(_) => return serde_json::to_string(&Vec::<SerializableMessage>::new()).unwrap(),
    };

    if let Some(thread) = threads.into_iter().find(|thread| thread.id == id) {
        let messages: Vec<SerializableMessage> = match serde_json::from_reader(messages_reader)
        {
            Ok(messages) => messages,
            Err(_) => return serde_json::to_string(&Vec::<SerializableMessage>::new()).unwrap(),
        };

        let result_messages: Vec<SerializableMessage> = thread.messages.into_iter()
            .filter_map(|message_id| {
                messages.iter().find(|message| message.id == message_id).cloned()
            })
            .collect();

        serde_json::to_string(&result_messages).unwrap()
    } else {
        serde_json::to_string(&Vec::<SerializableMessage>::new()).unwrap()
    }
}

async fn get_threads(folder: String, file_name: String) -> String {
    let path = Config::global()
        .out_dir
        .clone()
        .join(folder)
        .join(file_name);

    let file = match File::open(path) {
        Ok(file) => file,
        Err(_) => return serde_json::to_string(&Vec::<(String, Vec<u32>)>::new()).unwrap(),
    };
    let reader = BufReader::new(file);
    let messages: Vec<(String, Vec<u32>)> = match serde_json::from_reader(reader)
    {
        Ok(messages) => messages,
        Err(_) => return serde_json::to_string(&Vec::<(String, Vec<u32>)>::new()).unwrap(),
    };

    serde_json::to_string(&messages).unwrap()
}

pub async fn get_sorted_threads_by_date(folder: String) -> String {
    get_threads(folder, "date.json".to_string()).await
}

// Handlers
#[derive(Deserialize, Serialize)]
struct ErrorResponse {
    error: String,
}

#[derive(Deserialize)]
struct GetSortedThreadsByDateQuery {
    folder: String,
}

async fn sorted_threads_by_date_handle(query: GetSortedThreadsByDateQuery) -> Result<impl warp::Reply, warp::Rejection> {
    let result: Vec<(String, Vec<u32>)> = serde_json::from_str(&*get_sorted_threads_by_date(query.folder).await).unwrap();
    Ok(warp::reply::json(&result))
}

async fn get_folders_handle() -> Result<impl warp::Reply , warp::Rejection> {
    let result: Vec<String> = serde_json::from_str(&*get_folders().await).unwrap();
    Ok(warp::reply::json(&result))
}

#[derive(Deserialize)]
struct GetEmailQuery {
    id: String,
}

async fn get_email_handle(query: GetEmailQuery) -> Result<impl warp::Reply, warp::Rejection> {
    let result: String = get_email(query.id).await;
    Ok(warp::reply::html(result))
}

#[derive(Deserialize)]
struct GetThreadMessagesQuery {
    id: u32,
}

async fn get_thread_messages_handle(query: GetThreadMessagesQuery) -> Result<impl warp::Reply, warp::Rejection> {
    let result: Vec<SerializableMessage> = serde_json::from_str(&*get_thread_messages(query.id).await).unwrap();
    Ok(warp::reply::json(&result))
}

#[derive(Deserialize)]
struct CreateFolderRequest {
    name: String,
}

async fn create_folder_handle(data: CreateFolderRequest) -> Result<impl warp::Reply, warp::Rejection> {
    match create_folder_lar(data.name.clone()).await {
        Ok(_) => Ok(warp::reply::with_status(warp::reply::json(&"OK"), StatusCode::OK)),
        Err(_) => {
            let error_response = ErrorResponse {
                error: format!("Cannot create folder {}", data.name),
            };
            Ok(warp::reply::with_status(warp::reply::json(&error_response), StatusCode::BAD_REQUEST))
        }
    }
}

#[derive(Deserialize)]
struct DeleteFolderRequest {
    name: String,
}

async fn delete_folder_handle(data: DeleteFolderRequest) -> Result<impl warp::Reply, warp::Rejection> {
    match delete_folder_lar(data.name.clone()).await {
        Ok(_) => Ok(warp::reply::with_status(warp::reply::json(&"OK"), StatusCode::OK)),
        Err(_) => {
            let error_response = ErrorResponse {
                error: format!("Cannot delete folder {}", data.name),
            };
            Ok(warp::reply::with_status(warp::reply::json(&error_response), StatusCode::BAD_REQUEST))
        }
    }
}

#[derive(Deserialize)]
struct RenameFolderRequest {
    old_name: String,
    new_name: String,
}

async fn rename_folder_handle(data: RenameFolderRequest) -> Result<impl warp::Reply, warp::Rejection> {
    match rename_folder_lar(data.old_name.clone(), data.new_name.clone()).await {
        Ok(_) => Ok(warp::reply::with_status(warp::reply::json(&"OK"), StatusCode::OK)),
        Err(_) => {
            let error_response = ErrorResponse {
                error: format!("Cannot rename folder from {} to {}", data.old_name, data.new_name),
            };
            Ok(warp::reply::with_status(warp::reply::json(&error_response), StatusCode::BAD_REQUEST))
        }
    }
}

#[derive(Deserialize)]
struct SearchQuery {
    list: String,
    limit: i32,
    offset: i32,
    query: String
}

async fn search_handle(query: SearchQuery) -> Result<impl warp::Reply, warp::Rejection> {
    match SearchSonic::search_document("emails", &*query.list, &*query.query, query.limit, query.offset){
        Ok(result) => Ok(warp::reply::json(&result)),
        Err(_) => Ok(warp::reply::json(&Vec::<String>::new()))
    }
}